#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
//#include <WebSocketsClient.h>
#include <DHTesp.h>
#include <NTPClient.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"

// --------- Information to Connect to Adafruit IO ------------------------------------------------------
#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883                   // use 8883 for SSL
#define AIO_USERNAME    ""
#define AIO_KEY         ""
#define TEMP_FEED       "/feeds/"
#define HUM_FEED        "/feeds/"
#define WIND_FEED       "/feeds/"
#define RAF_FEED       "/feeds/"

// --------- Frequenza aggiornamenti --------------------------------------------------------------------
#define WIND_INTERVAL 5*1000 // Secondi
#define IOTTEMP_INTERVAL 2*61*1000 // Minuti
#define IOTWIND_INTERVAL 1*60*1000 // Minuti

// --------- PINOUT -------------------------------------------
#define pinTemp D1 // Pin DHT-11
#define pinVento D2 // Pin Anemometro

// ----------- Parametri WiFi --------------------------------------------------------------------------
#define MySSID "" // Change to your Wifi network SSID
#define MyWifiPassword "" // Change to your Wifi network password

#define ragAn 50 // Raggio anemometro in mm
#define filtro 0.99 // Filtro passa basso

// --------- Stringhe IOT -------------------------------------------------------------------------------
ESP8266WiFiMulti WiFiMulti;
//WebSocketsClient webSocket;
WiFiClient client;
Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);
Adafruit_MQTT_Publish FeedTemperatura = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME TEMP_FEED);
Adafruit_MQTT_Publish FeedUmidita = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME HUM_FEED);
Adafruit_MQTT_Publish FeedVento = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME WIND_FEED);
Adafruit_MQTT_Publish FeedRaffica = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME RAF_FEED);

//-------------------------------------------------------------------------------------------------------
// --------- Non toccare da qui in poi ------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------
float humidity;
float temperature;
float temperaturefh;
float wind;
float windmax;
float windavg;
uint64_t windPulses = 0;
DHTesp dht;
uint64_t iotTempUpdateTimestamp = 0;
uint64_t iotWindUpdateTimestamp = 0;
uint64_t windUpdateTimestamp = 0;

//-----------------------CONNESSIONE SERVIZIO MQTT--------------------------------------
void MQTT_connect() {
  int8_t ret;

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  } else {
    //Serial.println("MQTT Connessione fallita");
  }

  Serial.print("Connecting to MQTT... ");

  uint8_t retries = 3;
  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
    Serial.println(mqtt.connectErrorString(ret));
    Serial.println("Retrying MQTT connection in 5 seconds...");
    mqtt.disconnect();
    delay(5000);  // wait 5 seconds
    retries--;
    if (retries == 0) {
      // basically die and wait for WDT to reset me
      //while (1);
      return;
    }
  }
  Serial.println("MQTT Connected!");
}

//----------------------------LETTTURA TEMPERATURA--------------------------------------
void aggiornaTemperatura() {
  //Lettura temperatura da DHT
  //delay(dht.getMinimumSamplingPeriod());

  humidity = dht.getHumidity();
  temperature = dht.getTemperature();
  temperature = temperature - 2.0;
  temperaturefh = dht.toFahrenheit(temperature);

  Serial.print(dht.getStatusString());
  Serial.print("\t");
  Serial.print(humidity, 1);
  Serial.print("\t\t");
  Serial.print(temperature, 1);
  Serial.print("\t\t");
  Serial.print(dht.toFahrenheit(temperature), 1);
  Serial.print("\t\t");
  Serial.print(dht.computeHeatIndex(temperature, humidity, false), 1);
  Serial.print("\t\t");
  Serial.println(dht.computeHeatIndex(dht.toFahrenheit(temperature), humidity, true), 1);
}

ICACHE_RAM_ATTR void detectsMovement() {
  static unsigned long last_interrupt_time = 0;
  unsigned long interrupt_time = millis();
  
  if (interrupt_time - last_interrupt_time > 50)
  {
    ++windPulses;
  }
  last_interrupt_time = interrupt_time;
}

void setup() {
  Serial.begin(115200);
  Serial.println("Accensione...");

  dht.setup(pinTemp, DHTesp::DHT22); // Connect DHT sensor to GPIO

  WiFiMulti.addAP(MySSID, MyWifiPassword);
  Serial.println();
  Serial.print("Connecting to Wifi: ");
  Serial.println(MySSID);

  // Waiting for Wifi connect
  while (WiFiMulti.run() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");

  }
  if (WiFiMulti.run() == WL_CONNECTED) {
    Serial.println("");
    Serial.print("WiFi connected. ");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
  }
  delay(2000);

  ArduinoOTA.setHostname("ESP-Meteo");
  ArduinoOTA.onStart([]() {
    Serial.println("Start");
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
    else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
    else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
    else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
    else if (error == OTA_END_ERROR) Serial.println("End Failed");
  });
  ArduinoOTA.begin();
  Serial.print("OTA Update Ready");
  Serial.print(" IP address: ");
  Serial.println(WiFi.localIP());

  pinMode(pinVento, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(pinVento), detectsMovement, FALLING);
}

void loop() {
  uint64_t now = millis();

  ArduinoOTA.handle();

  if ((now - iotTempUpdateTimestamp) > IOTTEMP_INTERVAL) {
    aggiornaTemperatura();
    MQTT_connect();
    //delay(500);
    iotTempUpdateTimestamp = now;
    FeedTemperatura.publish(temperature);
    FeedUmidita.publish(humidity);
    Serial.println("[TEMP] Pubblicato");
    mqtt.disconnect();
  }
  if ((now - windUpdateTimestamp) > WIND_INTERVAL) {
    windUpdateTimestamp = now;
    wind = (2*ragAn*PI) * windPulses * (60.0 / WIND_INTERVAL); //millimetri al minuto
    wind = wind / 60000.0; //chilometri orari
    if (wind > windmax ) {
      windmax = wind;
    }
    windavg = (windavg * filtro) + (wind * (1.0 - filtro));
    windPulses = 0;
    //Serial.print("Impulsi: ");
    //Serial.println(int(windPulses));
  }
  if ((now - iotWindUpdateTimestamp) > IOTWIND_INTERVAL) {
    iotWindUpdateTimestamp = now;
    MQTT_connect();
    FeedVento.publish(windavg);
    FeedRaffica.publish(windmax);
    windmax = 0.0;
    //windavg = 0.0;
    Serial.println("[VENTO] Pubblicato");
    mqtt.disconnect();
  }
}
